/*
** Module   : datetime.java
** Abstract : Progress 4GL compatible datetime object
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------------
** 001 GES 20130128 Initial commit with basic functionality.
** 002 GES 20130322 Fixed the BDT c'tor and modified the assign(BDT,boolean) to be the
**                  common location for the _POLY morphing. This is safe because all current
**                  usage would have generated a ClassCastException if that method was ever used
**                  with an invalid type. Now, instead of the CCE, the code will morph the value
**                  as is possible. The morphing code still needs to be finished.
** 003 OM  20130422 Added int64 support.
*/

package com.goldencode.p2j.util;

import java.util.*;

/**
 * A class that represents a Progress 4GL compatible datetime object whose
 * data is mutable.  All Progress language features related to date data
 * types are supported including all function that can be accessed via the
 * date aware operators and built-in Progress date-related functions.
 * <p>
 * This class extends the <code>date</code> class so all methods should work
 * directly. If they don't they will be overridden. Also should the following
 * static implementations:
 * <p>
 * <pre>
 * +        operator (datetime, integer)              {@link #plusMillis(datetime, NumberType)} 
 * -        operator (datetime, integer)              {@link #minusMillis(datetime, NumberType)}
 * -        operator (datetime, datetime)             {@link #differenceNum}
 * string   function (default format)                 {@link #toString()}
 *                   (user-defined format)            {@link #toString(String)} 
 *                   (export format)                  {@link #toStringExport()}
 * now      function                                  {@link #datetime()}
 * datetime function (date)                           {@link #datetime(date)}
 * datetime function (date, integer)                  {@link #datetime(date, NumberType)}
 * datetime function (integer m, d, y, h, mm, s, ms)  {@link #datetime(NumberType, NumberType, NumberType, NumberType, NumberType, NumberType, NumberType)}
 * datetime function (string parameter)               {@link #datetime(String)}
 * </pre>
 * <p>
 * Default data type display formats: DATETIME 99/99/9999 HH:MM:SS.SSS
 *
 * @author    OM
 */
public class datetime
extends date
{
   /** The default format string. */
   private static final String defaultFormat = "99/99/9999 HH:MM:SS.sss";

   /** The milliseconds since midnight. */
   private int time = -1;

   /**
    * Default constructor which creates an instance that represents current
    * date and time using the default timezone.  This is the equivalent of
    * the NOW built-in function.
    */
   public datetime()
   {
      TimeZone zone = getDefaultTimeZone();
      
      setTimeZone(zone);
      
      // base class will initialize the date portion, here we just do
      // the time part
      Calendar now = Calendar.getInstance(zone);
      this.time = (int) millisSinceMidnight(now);
   }

   /**
    * This is a special c'tor which should be used only when converting the
    * value returned by a function or method with polymorphic return type into the
    * expected type (i.e. DYNAMIC-FUNCTION()).  In such cases, the 4GL does some
    * automatic type conversion (see {@link #assign(BaseDataType,boolean)}).
    * 
    * @param    value
    *           The value to be used for this instance.
    */
   public datetime(BaseDataType value)
   {
      if (value == null || value.isUnknown())
      {
         setUnknown();
      }
      else
      {
         assign(value, false);
      }
   }
   
   /**
    * Constructs an instance with the exact same state as the passed
    * instance.  This can be used both as a copy constructor as well as an
    * implementation of the DATETIME() built-in function when called with a
    * <code>date</code> subclass.
    *
    * @param    d
    *           The instance from which to copy state.
    */
   public datetime(date d)
   {
      super(d); // assign date part (and timezone)

      if (d instanceof datetime)
      {
         // then the time, if the case
         this.time = ((datetime)d).time;
      }
   }

   /**
    * Constructs an instance with the exact same state as the passed
    * instance and having the specified time. This is an implementation of the
    * DATETIME() built-in function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the default timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * 
    * @param    d
    *           The date from which to copy.
    * @param    time
    *           The milliseconds since midnight.
    */
   public datetime(date d, NumberType time)
   {
      if (d.isUnknown() || time.isUnknown())
      {
         setUnknown();
      }
      else
      {
         instantiateFromDateTime(d, time.intValue(), getDefaultTimeZone());
      }
   }

   /**
    * Constructs an instance with the exact same state as the passed
    * instance and having the specified time. This is an implementation of the
    * DATETIME() built-in function.
    * <p>
    * The time value (positive or negative) is added to the date and the
    * resulting value will be used to determine the date and time based on
    * the default timezone.  In other words, the date value can change in
    * either direction by an arbitrary magnitude.  This is how it is done in
    * the 4GL.
    * 
    * @param    d
    *           The date from which to copy.
    * @param    time
    *           The milliseconds since midnight.
    */
   public datetime(date d, double time)
   {
      this(d, new decimal(time));
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME() built-in function.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    */
   public datetime(double month,
                   double day,
                   double year,
                   double hours,
                   double minutes,
                   double seconds,
                   double millis)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime((int) hours, (int) minutes, (int) seconds, (int) millis);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME() built-in function.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    */
   public datetime(double month,
                   double day,
                   double year,
                   double hours,
                   double minutes,
                   double seconds)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime((int) hours, (int) minutes, (int) seconds, 0);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME() built-in function.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    */
   public datetime(double month,
                   double day,
                   double year,
                   double hours,
                   double minutes)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime((int) hours, (int) minutes, 0, 0);
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component of the date and with a time value of 0.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    */
   public datetime(double month, double day, double year)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime(0, 0, 0, 0);
   }
   
   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME() built-in function. Any
    * component that is <code>unknown</code> will result in the instance being
    * <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    */
   public datetime(NumberType month,
                   NumberType day,
                   NumberType year,
                   NumberType hours,
                   NumberType minutes,
                   NumberType seconds,
                   NumberType millis)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // use the version of setTime that handles unknown properly
      setTime(hours, minutes, seconds, millis);
   }

   /**
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME() built-in function. Any
    * component that is <code>unknown</code> will result in the instance being
    * <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    seconds
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    */
   public datetime(NumberType month,
                   NumberType day,
                   NumberType year,
                   NumberType hours,
                   NumberType minutes,
                   NumberType seconds)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // use the version of setTime that handles unknown properly
      setTime(hours, minutes, seconds, new int64(0));
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component.  This is a variant of the DATETIME() built-in function. Any
    * component that is <code>unknown</code> will result in the instance being
    * <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    minutes
    *           The minute of the hour from 0 to 59 inclusive.
    */
   public datetime(NumberType month,
                   NumberType day,
                   NumberType year,
                   NumberType hours,
                   NumberType minutes)
   {
      // the super class will take care of the date part
      super(month, day, year);

      // use the version of setTime that handles unknown properly
      setTime(hours, minutes, new int64(0), new int64(0));
   }

   /**               
    * Construct a datetime instance using explicit values for each discrete
    * component of the date and with a time value of 0. Any <code>unknown</code>
    * input will result in an instance that is <code>unknown</code>.
    * 
    * @param    month
    *           The month number from 1 to 12 inclusive.
    * @param    day
    *           The day of the month from 1 to the maximum days in that month,
    *           inclusive
    * @param    year
    *           The Progress 4GL year which may be a negative number which
    *           represents the BC era or a positive AD era year. The value must
    *           be between -32768 and 32767 and it cannot be 0 since the year
    *           0 is invalid in Progress. The value will be windowed into the
    *           right century if it is 2 digits.  It will be used unchanged if
    *           it is 3 or more digits.
    */
   public datetime(NumberType month, NumberType day, NumberType year)
   {
      // the super class will take care of the date part
      super(month, day, year);

      setTime(0, 0, 0, 0);
   }
   
   /**
    * Creates a new instance from the given date and time specification. This
    * is one form of the DATETIME built-in function. This is also the form of
    * the constructor that is used to handle initializer expressions (literals).
    * <p>
    * The specification must match the following format rules:
    * <p>
    * <ol>
    *   <li> Date value (this is not optional).  The string cannot start with
    *        whitespace. The first input is a date value that has at least a
    *        month and day specified or a month, day and year. The values
    *        are parsed in order as specified in SESSION:DATE-FORMAT. Zero
    *        padding is allowed but not required.
    *   <li> Date and time delimiter (not optional). Whitespace or a letter T
    *        (upper or lowercase) which is followed by whitespace. If the letter
    *        T is used, it cannot be preceded by whitespace. No other characters
    *        are allowed.
    *   <li> Time value (optional). If not specified, it is set to 0. It is
    *        specified in HH:MM:SS.sss format, however right side components
    *        can be left off (so HH, HH:MM, HH:MM:SS and HH:MM:SS.sss are all
    *        valid forms). Zero padding is allowed but not required, so hours
    *        can be H or HH, minutes M or MM, seconds S or SS and milliseconds
    *        can be s, ss or sss. Delimiters can only be there if the following
    *        component is included.
    *   <li> Whitespace (optional).
    *   <li> Timezone specification (optional). This is in the form of a sign
    *        (+ or -) then the hours (1 or 2 digits) a colon delimiter and then
    *        minites (1 or 2 digits). Zero padding is allowed but not required.
    *        The minutes can be left off but if so, then the delimiter must
    *        also be left off.  If specified, the instance will be set using
    *        the timezone specified EVEN THOUGH THE TIMEZONE OFFSET IS NOT
    *        STORED (since this is not a datetimetz).
    *   <li> Whitespace (optional).
    * </ol>
    *
    * @param    spec
    *           The date and time specification.
    */
   public datetime(String spec)
   throws ErrorConditionException
   {
      instantiateFromStringWorker(spec);
   }

   /**
    * Creates a new instance from the given date and time specification. This
    * is one form of the DATETIME built-in function. If the input is
    * <code>unknown</code> the instance will be <code>unknown</code>.
    * <p>
    * The specification must match the following format rules:
    * <p>
    * <ol>
    *   <li> Date value (this is not optional).  The string cannot start with
    *        whitespace. The first input is a date value that has at least a
    *        month and day specified or a month, day and year. The values
    *        are parsed in order as specified in SESSION:DATE-FORMAT. Zero
    *        padding is allowed but not required.
    *   <li> Date and time delimiter (not optional). Whitespace or a letter T
    *        (upper or lowercase) which is followed by whitespace. If the letter
    *        T is used, it cannot be preceded by whitespace. No other characters
    *        are allowed.
    *   <li> Time value (optional). If not specified, it is set to 0. It is
    *        specified in HH:MM:SS.sss format, however right side components
    *        can be left off (so HH, HH:MM, HH:MM:SS and HH:MM:SS.sss are all
    *        valid forms). Zero padding is allowed but not required, so hours
    *        can be H or HH, minutes M or MM, seconds S or SS and milliseconds
    *        can be s, ss or sss. Delimiters can only be there if the following
    *        component is included.
    *   <li> Whitespace (optional).
    *   <li> Timezone specification (optional). This is in the form of a sign
    *        (+ or -) then the hours (1 or 2 digits) a colon delimiter and then
    *        minites (1 or 2 digits). Zero padding is allowed but not required.
    *        The minutes can be left off but if so, then the delimiter must
    *        also be left off.  If specified, the instance will be set using
    *        the timezone specified EVEN THOUGH THE TIMEZONE OFFSET IS NOT
    *        STORED (since this is not a datetimetz).
    *   <li> Whitespace (optional).
    * </ol>
    *
    * @param    spec
    *           The date and time specification.
    */
   public datetime(character spec)
   throws ErrorConditionException
   {
      if (spec.isUnknown())
      {
         setUnknown();
      }
      else
      {
         instantiateFromStringWorker(spec.toStringMessage());
      }
   }
   
   /**
    * Creates an instance from a <code>java.util.Date</code> instance.
    * <p>
    * If a timezone is specified, it is stored to be used when instantiating
    * <code>java.util.Date</code> instances using this instance's data.    
    *
    * @param    d
    *           The <code>Date</code>instance from which to pattern the
    *           instance or <code>null</code> to obtain the current time.
    * @param    zone
    *           The timezone whose offset is to be used to calculate the 
    *           date or <code>null</code> if the default JVM timezone is
    *           to be used.
    */
   public datetime(Date d, TimeZone zone)
   {
      if (d == null)
      {
         setUnknown();
      }
      else
      {
         zone = (zone == null) ? getDefaultTimeZone() : zone;
         
         long millis = dateToLocalMillis(d, zone);
         long julian = millisToJulianDay(millis);
         long remain = julianDayToMillis(julian);
         
         instantiateFromDateTime(new date(julian), (int) remain, zone);
      }
   }

   /**
    * Returns the number of milliseconds difference between the datetimes of 
    * two instances.
    *
    * @param    op1
    *           The left operand.
    * @param    op2
    *           The right operand (instance to subtract from the left
    *           operand).
    *
    * @return   The difference in days between the two datetimes or 
    *           <code>unknown value</code> if any input is
    *           <code>unknown value</code>.
    */
   public static int64 differenceNum(datetime op1, datetime op2)
   {
      if (op1.isUnknown() || op2.isUnknown())
      {
         return new int64();
      }
      
      return new int64(op1.getAbsoluteTimeOffset() -
                       op2.getAbsoluteTimeOffset());
   }
   
   /**
    * Obtain the millis since midnight for this instance. This is an 
    * implementation of the <code>TIME()</code> built-in function.
    * 
    * @return   The milliseconds since midnight.
    */
   public int64 getTimeNum()
   {
      return isUnknown() ? new int64() : new int64(time);
   }

   /**
    * Obtain the millis from midnight.
    * 
    * @return  milliseconds passed from midnight.
    */
   public int getTime()
   {
      return time;
   }

   /**
    * Sets the millis from midnight.
    * 
    * @param   millis
    *          Milliseconds passed from midnight.
    */
   public void setTime(NumberType millis)
   {
      if (millis.isUnknown())
      {
         setUnknown();
      }
      else 
      {
         setTime(millis.longValue());
      }
   }

   /**
    * Sets the milliseconds since midnight.
    * 
    * @param    millis
    *           Milliseconds since midnight.
    */
   public void setTime(long millis)
   {
      // TODO: check what happens when millis is negative or more than a day
      time = (int) millis;
   }

   /**
    * Extract the hours time-element from this datetime.
    * 
    * @return   The hour of this datetime object.
    */
   public int getHours()
   {
      return time / (1000 * 60 * 60);
   }

   /**
    * Extract the hours time-element from this datetime.
    * 
    * @return  the hour of this datetime object or unknown value if this
    *          is an unknown object.
    */
   public int64 getHoursNum()
   {
      return isUnknown() ? new int64() : new int64(getHours());
   }

   /**
    * Extract the minutes time-element from this datetime.
    * 
    * @return the minute of this datetime object
    */
   public int getMinutes()
   {
      int timeInMinutes = time / (1000 * 60);
      return timeInMinutes % 60;
   }

   /**
    * Extract the minutes time-element from this datetime.
    * 
    * @return  the minute of this datetime object or unknown value if this
    *          is an unknown object.
    */
   public int64 getMinutesNum()
   {
      return isUnknown() ? new int64() : new int64(getMinutes());
   }

   /**
    * Extract the seconds time-element from this datetime.
    * 
    * @return the second of this datetime object
    */
   public int getSeconds()
   {
      int timeInSeconds = time / 1000;
      return timeInSeconds % 60;
   }

   /**
    * Extract the seconds time-element from this datetime.
    * 
    * @return  the second of this datetime object or unknown value if this
    *          is an unknown object.
    */
   public int64 getSecondsNum()
   {
      return isUnknown() ? new int64() : new int64(getSeconds());
   }

   /**
    * Extract the milliseconds time-element from this datetime.
    * 
    * @return the millisecond of this datetime object
    */
   public int getMilliseconds()
   {
      return time % 1000;
   }

   /**
    * Extract the milliseconds time-element from this datetime.
    * 
    * @return  the millisecond of this datetime object or unknown value if 
    *          this is an unknown object.
    */
   public int64 getMillisecondsNum()
   {
      return isUnknown() ? new int64() : new int64(getMillisecondsNum());
   }

   /**
    * Sets the internal time. This method will validate the parameters and
    * raise an error if they are incorrect.Any input that is
    * <code>unknown</code> will cause this instance to be <code>unknown</code>.
    * 
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    mins
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    secs
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    */
   public void setTime(NumberType hours,
                       NumberType mins,
                       NumberType secs,
                       NumberType millis)
   {
      if (hours.isUnknown() ||
          mins.isUnknown()  ||
          secs.isUnknown()  ||
          millis.isUnknown())
      {
         setUnknown();
      }
      else
      {
         setTime(hours.longValue(),
                 mins.longValue(),
                 secs.longValue(),
                 millis.longValue());
      }
   }

   /**
    * Sets the internal time. This method will validate the parameters and
    * raise an error if they are incorrect.
    * 
    * @param    hours
    *           The hour of the day from 0 to 23 inclusive.
    * @param    mins
    *           The minute of the hour from 0 to 59 inclusive.
    * @param    secs
    *           The seconds of the minute from 0 to 62 inclusive. This accounts
    *           for leap seconds.  Please note that Progress documents 61 as
    *           the limit. It turns out that 62 is the real limit.
    * @param    millis
    *           The milliseconds of the second from 0 to 999 inclusive.
    */
   public void setTime(long hours, long mins, long secs, long millis)
   {
      if (hours < 0 || hours > 23)
      {
         ErrorManager.recordOrThrowError(11260, "Invalid hours in datetime value");
         return;
      }

      if (mins < 0 || mins > 59)
      {
         ErrorManager.recordOrThrowError(11256, "Invalid minutes in datetime value");
         return;
      }

      if (secs < 0 || secs > 62)
      {
         ErrorManager.recordOrThrowError(11253, "Invalid seconds in datetime value");
         return;
      }

      if (millis < 0 || millis > 999)
      {
         ErrorManager.recordOrThrowError(11252, "Invalid milliseconds in datetime value");
         return;
      }

      time = (int) (millis + 1000 * (secs + 60 * (mins + 60 * hours)));
   }

   /**
    * Adds <code>milliseconds</code> to the datetime of the left operand and 
    * returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to add. 
    *          Note that this can be a negative number to effectively 
    *          subtract instead of add.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if <code>dt</code> 
    *          is <code>unknown value</code>.
    */
   public static datetime plusMillis(datetime dt, long millis)
   {
      if (dt.isUnknown())
      {
         return datetime.instantiateUnknownDatetime();
      }
      long newMillis = dt.time + millis;
      long days = dt.longValue();
      if (newMillis < 0)
      {
         // must go back in time!
         // TODO: OM: here there is a bug, must be fixed
         days -= - newMillis / date.MILLIS_PER_DAY;
         newMillis = - newMillis % date.MILLIS_PER_DAY;
      }
      else if (newMillis >= date.MILLIS_PER_DAY)
      {
         // go forward in time
         days += newMillis / date.MILLIS_PER_DAY;
         newMillis = newMillis % date.MILLIS_PER_DAY;
      }
      return new datetime(new date(days), (int)newMillis);
   }

   /**
    * Adds <code>milliseconds</code> to the datetime of the left operand and 
    * returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to add. 
    *          Note that this can be a negative number to effectively 
    *          subtract instead of add.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if any input is
    *          <code>unknown value</code>.
    */
   public static datetime plusMillis(datetime dt, NumberType millis)
   {
      return (dt.isUnknown() || millis.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, millis.longValue());
   }

   /**
    * Subtracts <code>milliseconds</code> to the datetime of the left operand
    * and returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to subtract. 
    *          Note that this can be a negative number to effectively 
    *          add instead of subtract.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if <code>dt</code> 
    *          is <code>unknown value</code>.
    */
   public static datetime minusMillis(datetime dt, long millis)
   {
      return (dt.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, -millis);
   }

   /**
    * Subtracts <code>milliseconds</code> to the datetime of the left operand
    * and returns a new instance representing that datetime.
    *
    * @param   dt
    *          The left operand.
    * @param   millis
    *          The right operand (the number of milliseconds to subtract. 
    *          Note that this can be a negative number to effectively 
    *          add instead of subtract.
    *
    * @return  The date which is <code>millis</code> different from the
    *          left operand or <code>unknown value</code> if any input is
    *          <code>unknown value</code>.
    */
   public static datetime minusMillis(datetime dt, NumberType millis)
   {
      return (dt.isUnknown() || millis.isUnknown()) ?
              datetime.instantiateUnknownDatetime() :
              plusMillis(dt, -millis.longValue());
   }

   /**
    * Returns the time in milliseconds (MTIME, similar to TIME, which 
    * returns seconds since midnight).
    * 
    * @return  the time in milliseconds.
    */
   public static int64 millisecondsSinceMidnight()
   {
      return new int64(new datetime().time);
   }

   /**
    * Returns the time in milliseconds (MTIME, similar to TIME, which 
    * returns seconds since midnight).
    * 
    * @param   dt
    *          the datetime or datetime-tz to evaluate.
    *
    * @return  the time in milliseconds.
    */
   public static int64 millisecondsSinceMidnight(datetime dt)
   {
      return dt.isUnknown() ? new int64() : new int64(dt.time);
   }

   /**
    * Creates a new <code>datetime</code> instance that represents the 
    * <code>unknown value</code>.
    * 
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public static datetime instantiateUnknownDatetime()
   {
      datetime ret = new datetime();
      ret.setUnknown();
      return ret;
   }

   /**
    * Creates a new instance of the same type that represents the 
    * <code>unknown value</code>.
    *
    * @return   An instance that represents the <code>unknown value</code>.
    */
   public BaseDataType instantiateUnknown()
   {
      return instantiateUnknownDatetime();
   }

   /**
    * Creates a new instance of the same type that represents the 
    * default initialized value.
    *
    * @return   An instance that represents the default value.
    */
   public BaseDataType instantiateDefault()
   {
      return instantiateUnknownDatetime();
   }

   /**
    * Return the default display format string for this type.
    *
    * @return   The default format string.
    */
   public String defaultFormatString()
   {
      return defaultFormat;
   }   

   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    * <p>
    * If the value is not of type <code>datetime</code>, the following automatic type
    * conversion will occur:
    * <p>
    * <pre>
    * TODO: implement and document this
    * </pre>
    * <p>
    * This conversion is meant to handle the cases of built-in functions and methods
    * in the 4GL which have polymorphic return types (e.g. DYNAMIC-FUNCTION()).
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(BaseDataType value, boolean force)
   {
      if (value instanceof datetime)
      {
         assign((datetime) value, force);
      }
      else
      {
         // TODO: implement type conversion here for _POLY cases 
      }
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    * @param    force
    *           Flag to indicate if a pending error should be ignored (see 
    *           {@link ErrorManager#isPendingError} and the assignment
    *           forced regardless.  Note that normal assignment silently
    *           exits with no change if there is a pending error. If this
    *           flag is <code>true</code>, that check is bypassed and the
    *           assignment will always be done.
    */
   public void assign(datetime value, boolean force)
   {
      if (force || !ErrorManager.isPendingError())
      {
         time = value.time;
      }
      super.assign((date) value, force);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance. The configuration (time-zone) of this
    * instance is not set.  Use {@link #deepAssign} if both state and
    * configuration must be set.
    * <p>
    * This is a "no operation" in the case where silent error mode is active
    * AND an error has occurred in the current processing since the opening
    * call to {@link ErrorManager#silentErrorEnable}. This protection is
    * needed to ensure that the state of this instance does not change in
    * this case.  This will not affect undo because that only occurs when
    * silent error mode is disabled.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(datetime value)
   {
      assign(value, false);
   }
   
   /**
    * Sets the state (data and unknown value) of this instance based on the
    * state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void assign(Undoable value)
   {
      assign((datetime) value, false);
   }
   
   /**
    * Sets the state (data and unknown value) and configuration (time-zone)
    * of this instance based on the state of the passed instance.
    *
    * @param    value
    *           The instance from which to copy state.
    */
   public void deepAssign(datetime value)
   {
      super.deepAssign(value);
      time = value.time;
   }
   
   /**
    * Compares this instance with the specified instance and returns a -1
    * if this instance is less than the specified, 0 if the two instances
    * are equal and 1 if this instance is greater than the specified
    * instance.  This is the implementation of the <code>Comparable</code>
    * interface.
    * <p>
    * The algorithm will fail to give meaningful results in the case where
    * one tries to sort against other objects that do not represent compatible
    * values.
    *
    * @param    obj
    *           The date instance to compare against.
    *
    * @return   -1, 0 or 1 depending on if this instance is less than,
    *           equal to or greater (respectively) than the specified
    *           instance <code>obj</code>.
    */
   public int compareTo(Object obj)
   {
      date that = null;
      
      if (obj instanceof date)
      {
         that = (date) obj;
      }
      else
      {
         return -1;
      }

      if (this.isUnknown() || that.isUnknown())
      {
         //TODO: decide if two unknowns are equals
         return -1;
      }

      int result = 0;
      long thisAbs = this.getAbsoluteTimeOffset();
      long thatAbs = that.getAbsoluteTimeOffset();
      
      if (thisAbs < thatAbs)
      {
         result = -1;
      }
      else if (thisAbs > thatAbs)
      {
         result = 1;
      }

      return result;
   }

   /**
    * An equality test which handles the most common cases with the least
    * amount of overhead.  The superclass implementation is invoked otherwise.
    * <p>
    * This implementation is consistent with {@link #hashCode()}.
    * 
    * @see com.goldencode.p2j.util.BaseDataType#equals(java.lang.Object)
    */
   public boolean equals(Object o)
   {
      if (this == o)
      {
         return true;
      }
      
      if (o instanceof datetime)
      {
         datetime that = (datetime) o;
         
         boolean u1 = this.isUnknown();
         boolean u2 = that.isUnknown();
         
         if (u1 && u2)
         {
            return true;
         }
         
         if (u1 || u2)
         {
            return false;
         }

         return (this.longValue() == that.longValue()) && 
                (this.time == that.time);
      }
      
      return super.equals(o);
   }

   /**
    * Hash code implementation which is consistent with {@link
    * BaseDataType#equals}.
    *
    * @return  Hash code value for this object instance.
    */
   public int hashCode()
   {
      int result = 17;
      if (isUnknown())
      {
         return result;
      }
      
      result = 37 * result + (int) (longValue() ^ (longValue() >>> 32));
      result = 37 * result + time;
      
      return result;
   }

   /**
    * Does the same as standard <code>clone()</code> method but returns an
    * instance of <code>BaseDataType</code> and doesn't throw the
    * <code>CloneNotSupportedException</code>.
    *
    * @return   A clone of this instance.  
    */
   public BaseDataType duplicate()
   {
      return new datetime(this);
   }

   /**
    * Returns the current instance as an instance of the <code>Date</code>
    * class which will accurately represent the UTC time that corresponds
    * to the current date in the timezone stored in this instance or
    * in the default timezone if the stored timezone is <code>null</code>.
    * <p>
    * <strong>WARNING:</strong> since the J2SE <code>Date</code> class
    * implicitly encodes locale-specific time information along with date
    * information, the <code>Date</code> object returned by this method is
    * sensitive to the environment in which it is <em>interpreted</em>
    * downstream.  It is imperative therefore, when using the returned
    * <code>Date</code> in downstream code, that the timezone of such code
    * (e.g., a date formatter or <code>Calendar</code> instance) matches the
    * timezone stored in this <code>date</code> instance (if the stored
    * timezone is <code>null</code>, the JVM's default timezone should be used
    * in downstream code).  Otherwise, the discrepancy in timezones (including
    * any Daylight Savings offset) may cause downstream code to interpret the
    * date information in the returned object incorrectly, typically as one
    * day earlier than expected.
    *
    * @return   The date converted to be specific to the stored timezone or
    *           to the default timezone if the stored timezone is 
    *           <code>null</code>.
    */
   public Date dateValue()
   {
      return dateValue(getTimeZone());
   }
      
   /**
    * Returns the current instance as an instance of the <code>Date</code>
    * class which will accurately represent the UTC time that corresponds
    * to the current date in the timezone passed as a parameter or in the
    * default timezone if the passed timezone is <code>null</code>.
    * <p>
    * <strong>WARNING:</strong> since the J2SE <code>Date</code> class
    * implicitly encodes locale-specific time information along with date
    * information, the <code>Date</code> object returned by this method is
    * sensitive to the environment in which it is <em>interpreted</em>
    * downstream.  It is imperative therefore, when using the returned
    * <code>Date</code> in downstream code, that the timezone of such code
    * (e.g., a date formatter or <code>Calendar</code> instance) matches the
    * the <code>override</code> parameter to this method (if
    * <code>override</code> is <code>null</code>, the JVM's default timezone
    * should be used in downstream code).  Otherwise, the discrepancy in
    * timezones (including any Daylight Savings offset) may cause downstream
    * code to interpret the date information in the returned object
    * incorrectly, typically as one day earlier than expected.
    *
    * @param    override
    *           The timezone to use for interpreting this date or 
    *           <code>null</code> if the default JVM timezone should be used.
    *
    * @return   The date converted to be specific to the passed timezone.
    */
   public Date dateValue(TimeZone override)
   {
      if (override == null)
      {
         override = TimeZone.getDefault();
      }
      
      long total = longValue() + time;
      
      return neutralMillisToDate(julianDayToMillis(total), override);
   }
      
   /**
    * Obtains the text form of this date instance in ISO 8601 standard format.
    * 
    * @return   A string that is the ISO representation of this date
    *           ("YYYY-MM-DDTHH:MM:SS.SSS") or "?" if <code>unknown</code>.
    */
   protected String getIsoDate()
   {
      if (isUnknown())
      {
         return "?";
      }
      
      StringBuilder sb = new StringBuilder(super.getIsoDate());
      sb.append("T");
      sb.append(date.formatNDigits(getHours(), 2));
      sb.append(":");
      sb.append(date.formatNDigits(getMinutes(), 2));
      sb.append(":");
      sb.append(date.formatNDigits(getSeconds(), 2));
      sb.append(".");
      sb.append(date.formatNDigits(getMilliseconds(), 3));

      return sb.toString();
   }

   /**
    * Returns the absolute time in milliseconds since Java 'Epoch' or 0
    * if this is unknown.
    *
    * @return   The absolute time in milliseconds since the Java epoch.
    */
   @Override
   protected long getAbsoluteTimeOffset()
   {
      return isUnknown() ? 0 : (super.getAbsoluteTimeOffset() + time);
   }
   
   /**
    * Worker to set the date (in the superclass) and the time (in this class)
    * given all the basic inputs.
    *
    * @param    d
    *           The date instance to copy from.
    * @param    time
    *           Milliseconds since midnight.
    * @param    tz
    *           The effective timezone to use.
    */
   protected void instantiateFromDateTime(date d, int time, TimeZone tz)
   {
      long millis = julianDayToMillis(d.longValue()) + time;
      
      GregorianCalendar gc = neutralMillisToCalendar(millis, tz);
      
      setDayNumber(millisToJulianDay(calendarToLocalMillis(gc)));
      setTimeZone(tz);
      
      this.time = (int) millisSinceMidnight(gc);
   }
   
   /**
    * Parses the given date and time specification and assigns the values into
    * the instance.  The result will be a date and time value that is "local"
    * to the default timezone.  If a different timezone is specified in the
    * input, then the date and time values that are specified will be translated
    * into values "local" to the default timezone, before assignment.
    * <p>
    * The specification must match the following format rules:
    * <p>
    * <ol>
    *   <li> Date value (this is not optional).  The string cannot start with
    *        whitespace. The first input is a date value that has at least a
    *        month and day specified or a month, day and year. The values
    *        are parsed in order as specified in SESSION:DATE-FORMAT. Zero
    *        padding is allowed but not required.
    *   <li> Date and time delimiter (not optional). Whitespace or a letter T
    *        (upper or lowercase) which is followed by whitespace. If the letter
    *        T is used, it cannot be preceded by whitespace. No other characters
    *        are allowed.
    *   <li> Time value (optional). If not specified, it is set to 0. It is
    *        specified in HH:MM:SS.sss format, however right side components
    *        can be left off (so HH, HH:MM, HH:MM:SS and HH:MM:SS.sss are all
    *        valid forms). Zero padding is allowed but not required, so hours
    *        can be H or HH, minutes M or MM, seconds S or SS and milliseconds
    *        can be s, ss or sss. Delimiters can only be there if the following
    *        component is included.
    *   <li> Whitespace (optional).
    *   <li> Timezone specification (optional). This is in the form of a sign
    *        (+ or -) then the hours (1 or 2 digits) a colon delimiter and then
    *        minites (1 or 2 digits). Zero padding is allowed but not required.
    *        The minutes can be left off but if so, then the delimiter must
    *        also be left off.  If specified, the instance will be set using
    *        the timezone specified EVEN THOUGH THE TIMEZONE OFFSET IS NOT
    *        STORED (since this is not a datetimetz).
    *   <li> Whitespace (optional).
    * </ol>
    *
    * @param    spec
    *           The date and time specification.
    */
   @Override
   protected void instantiateFromStringWorker(String spec)
   {
      // TODO: implement + make sure all the errors are thrown properly!
      
      // if a TZ spec is used, then the other portions of the date and time
      // may change before being assigned; this may require the 
      // date.parseWorker() code to be refactored into something that just
      // does the parsing and error handling but not the call to
      // instantiateDate()
      
      // in other cases (no TZ spec):
      
      // date part can use date.instantiateFromStringWorker() after the date
      // has been split off
      
      // this code must parse the rest into pieces and implement the proper
      // calc, including honoring the timezone stuff when calculating the
      // millis
      
      // timezone spec can use date.specToOffset()
   }
   
   /**
    * Extract the portion of the calendar representing the number of
    * milliseconds since midnight.
    *
    * @param    cal
    *           The calendar instance to query.
    *
    * @return   The milliseconds since midnight.
    */
   private long millisSinceMidnight(Calendar cal)
   {
      int hour   = cal.get(Calendar.HOUR_OF_DAY);
      int mins   = cal.get(Calendar.MINUTE);
      int secs   = cal.get(Calendar.SECOND);
      int millis = cal.get(Calendar.MILLISECOND); 
      
      return (((((hour * 60) + mins) * 60) + secs) * 1000) + millis;
   }
}

